var util = require('util');
util._extend(util, require('core-util-is'));

var MPX = require('./core'),
    url = require('url'),
    qs = require('qs'),
    protocols = { http: require('http'), https: require('https') },
    validMethods = ['GET', 'POST', 'PATCH', 'DELETE'];

/**
 * The Request class stands behind all comunication to external HTTP services.
 * Its constructor accepts option that configures the Request instance to
 * such extent, that only the execute method with callback is only necessary
 * to call after.
 *
 * @example Request setup
 *   var request = new MPX.Request({
 *     host: 'ah.mediapeers.net',
 *     method: 'POST',
 *     path: '/assets',
 *     data: {
 *       asset: {
 *         key: '/deliveries/ftp/ASF2342ASFD/james_bond.jpg'
 *         // ommited
 *       }
 *     }
 *   });
 *   request.execute(function(error, response) {});
 *
 *  @param options [Object] a map that contains Request settings
 *  @option options protocol [String] protocol to use (supported are http and https)
 *  @option options hostname [String] destination host
 *  @option options path [String] destination path
 *  @option options method [String] HTTP method to use (supported are GET, POST, PATCH and DELETE)
 *  @option options data [Object] data to send to API
 *
 * @raise [Error] if unsupported protocl is given
 * @raise [Error] if hostname is empty
 * @raise [Error] if path is empty
 * @raise [Error] if unsupported HTTP method is given
 */
function Request(options) {
  if(util.isUndefined(options)) options = {};

  this.protocol = options.protocol || 'http';
  this.hostname = options.hostname;
  this.path = options.path;
  this.method = options.method;
  this.data = options.data || {};

  if(String(this.protocol).match(/^http(s?)$/) === null) {
    throw Error('Protocol "' + this.protocol +'" is not valid. Valid protocols are http,https');
  }

  if(util.isUndefined(this.hostname)) throw Error('Hostname must not be empty');
  if(util.isUndefined(this.path))     throw Error('Path must not be empty');

  if(validMethods.indexOf(this.method) === -1) {
    throw Error('Method "' + this.method + '" is not valid. Valid methods are ' + validMethods.toString());
  }

  this.urlEncoded = ['POST', 'PATCH'].indexOf(this.method) >= 0;
}


/**
 * Creates an Error based on received HTTP status code. If
 * status code is unknown, it return an unknown error.
 *
 * @example Error generation
 *   Request.createError(406);
 *   // => { [InvalidDataFormat: Invalid Data Format] name: 'InvalidDataFormat', statusCode: 406 }
 *
 * @param errorCode [Integer] HTTP Status Code
 * @return [Object] an instance of Error
 */
Request.createError = function(errorCode) {
  var errorCodes = {
    400: 'Bad Request',
    401: 'Not Authenticated',
    403: 'Action Not Authorized',
    404: 'Resource Not Found',
    406: 'Invalid Data Format',
    500: 'Internal Server Error',
    502: 'Bad Gateway',
    503: 'Service Unavailable'
  };

  var message = errorCodes[errorCode];
  if(util.isUndefined(message)) message = 'Unknown';

  var error = new Error(message);
  error.name = message.split(' ').join('');
  error.statusCode = errorCode;

  return error;
};

/**
 * Executes request that calls service API
 * @callback callback function(error, response)
 *   @param error [Error] the error given by response status code
 *   @param data [Object] data returned from the server
 *   @param response [Object] HTTP response
 */
Request.prototype.execute = function(callback) {
  if(!util.isFunction(callback)) throw Error('Callback is not a function');

  var data = '';

  var request = protocols[this.protocol].request(this.getOptions(), function(response) {
    response.setEncoding('utf8');

    response.on('data', function(chunk) {
      data += chunk;
    });

    response.on('end', function() {
      if(response.statusCode >= 400 && response.statusCode !== 422) {
        callback(Request.createError(response.statusCode));
      } else {
        var parsedData;

        try {
          parsedData = JSON.parse(data);
        }
        catch(e) {
          parsedData = {};
        }

        callback(null, parsedData, response);
      }
    });
  });

  request.on('error', function(error) {
    callback(error);
  });

  if(this.urlEncoded) request.write(this.data);
  request.end();
};

/**
 * Generates HTTP request options based on the HTTP request type.
 * For GET and DELETE requests, it encodes parameters directly
 * in the URL, however for POST and PATCH, it omits parameters
 * completely. Only format (this is always json) stays in URL.
 * It also sets respective headers like Content-Type and Content-Length
 * for POST and PATCH.
 *
 * @return [Object] a map of options for http(s).Request
 */
Request.prototype.getOptions = function() {
  var options = {
    hostname: this.hostname,
    port: (this.protocol === 'http' ? 80 : 443),
    path: {
      pathname: this.path,
      query: {
        format: 'json'
      }
    },
    method: this.method,
    headers: {
      'User-Agent': 'mediapeers MPX Node.js SDK'
    }
  };

  if(this.urlEncoded) {
    this.data = qs.stringify(this.data);
    options.headers['Content-Type'] = 'application/x-www-form-urlencoded';
    options.headers['Content-Length'] = Buffer.byteLength(this.data);
  } else {
    util._extend(options.path.query, this.data);
  }

  options.path = url.format(options.path);

  return options;
};

MPX.Request = Request;
