var util = require('util');
util._extend(util, require('core-util-is'));

var path = require('path'),
    fs = require('fs'),
    changeCase = require('change-case'),
    MPX = require('./core');

/**
 * Headers to be exposed in the response object
 */
var allowedHeaders = ['x-total-entries', 'x-total-pages', 'x-page', 'x-per-page', 'x-request-id'];

/**
 * Base class for services. The constructor validates prerequisites like endpoint and
 * version. If they miss, it throw an error. It then defines the service by calling
 * the defineService method.
 *
 * @abstract
 *
 * @example Initialize Asset Host instance
 *  var assetHost = new MPX.AssetHost({
 *    accessKey: 'access-key',
 *    secretKey: 'secret-key
 *  });
 *
 * @param [Object] config the configuration options
 * @option config endpoint [String] the host address on which API will be called
 * @option config version [String] the version of API to be used
 * @option config accessKey [String] your service access key
 * @option config secretKey [String] your service secret key
 *
 * @raise [Error] if endpoint is null
 * @raise [Error] if API version is null
 */
function Service(config) {
  if(util.isUndefined(config)) config = {};

  this.config = new MPX.Config(MPX.config);
  this.config.update(config);

  if(util.isNull(this.config.endpoint)) throw Error('Endpoint is not defined');
  if(util.isNull(this.config.version))  throw Error('API version is not defined');

  this.defineService();
}

/**
 * Loads service definition based on the constructor name and the version defined in
 * the configuration. For each operation in the definition calls the defineOperation method.
 *
 * @api private
 */
Service.prototype.defineService = function() {
  if(this.constructor.name !== 'Service') {
    var filename = changeCase.snakeCase(this.constructor.name) + '-' + this.config.version + '.json',
        definitionPath = path.join(__dirname, 'services', 'api', filename),
        serviceDefinition = JSON.parse(fs.readFileSync(definitionPath));

    if(serviceDefinition.version !== this.config.version) throw Error('Service definition does not match!');
    for(var operationName in serviceDefinition.operations) {
      this.defineOperation(operationName, serviceDefinition.operations[operationName]);
    }
  }
};

/**
 * Defines individual API services.
 *
 * @api private
 *
 * @example a definition of a new operation
 *  var definition = {
 *    http: {
 *      method: 'GET',
 *      path: '/asset/{assetId}'
 *    },
 *    input: {
 *      assetId: {
 *        required: true
 *      }
 *    }
 *  };
 *
 *  service.defineOperation('getAsset', definition);
 *  service.getAsset({ assetId: 123 }, function(error, response) { whatever; });
 *
 * @param name [String] the name of operation to be defined
 * @param definition [Object] that defines the structure of params to be
 *   sent to API as well as the HTTP method and path
 *   * http [Object] contains HTTP settings for request
 *     * method [String] HTTP method to be used (only GET, POST, PATCH and DELETE)
 *       are supported
 *     * path [String] a path to resource
 *   * input [Object] params definition
 *
 * @return [Function] a function that has been defined based
 *   on given definition
 */
Service.prototype.defineOperation = function(name, definition) {
  var service = this;
  service[name] = function(input, callback) {
    if(util.isFunction(input)) {
      callback = input;
      input = {};
    }

    if(util.isUndefined(input)) input = {};

    var options = {
      hostname: service.config.endpoint,
      method: definition.http.method,
      path: path.join('/', service.config.version, service.inputToURL(definition.http.path, input)),
      data: service.sanitizeInput(input, definition.input)
    };

    var request = new MPX.Request(options);
    request.execute(function(error, data, response) {
      if(error) {
        callback(error);
      } else {
        var result = {
          statusCode: response.statusCode,
          data: service.sanitizeOutput(data)
        };

        util._extend(result, service.sanitizeHeaders(response.headers));
        callback(null, result);
      }
    });
  };

 return service[name];
};

/**
 * Recursively converts JavaScript defined object with camelCase
 * keys to object that contains snake_case keys.
 *
 * @api private
 */
Service.prototype.sanitizeInput = function(input, definition) {
  if(definition.allowAny) return input;

  var sanitizedInput = {};

  for(var key in definition) {
    var value = input[key];
    if(util.isObject(value)) {
      sanitizedInput[changeCase.snakeCase(key)] = this.sanitizeInput(value, definition[key]);
    } else {
      if(util.isUndefined(value) && definition[key].required) {
        throw Error(key + ' is a required property');
      }
      else if(!util.isUndefined(value)) {
        sanitizedInput[changeCase.snakeCase(key)] = value;
      }
    }
  }

  return sanitizedInput;
};

/**
 * Converts all headers from param-case to camelCase. From custom
 * headers removes the x- prefix.
 *
 * @api private
 */
Service.prototype.sanitizeHeaders = function(headers) {
  var sanitizedHeaders = {};

  for(var i = 0; i < allowedHeaders.length; i++) {
    var allowedHeader = allowedHeaders[i];
    var value = headers[allowedHeader];

    if(!util.isUndefined(value)) sanitizedHeaders[changeCase.camelCase(allowedHeader.replace(/(^x)/, ''))] = value;
  }

  return sanitizedHeaders;
};

/**
 * Recursively converts Rails response defined object with
 * snake_case keys to objects that contain camelCase keys.
 *
 * @api private
 */
Service.prototype.sanitizeOutput = function(output) {
  var service = this,
      sanitizedOutput;

  if(util.isArray(output)) {
    sanitizedOutput = [];

    for(var i = 0; i < output.length; i++) {
      sanitizedOutput[i] = service.sanitizeOutput(output[i]);
    }
  } else {
    sanitizedOutput = {};
    for(var key in output) {
      sanitizedOutput[changeCase.camelCase(key)] = output[key];
    }
  }

  return sanitizedOutput;
};

/**
 * Replaces variables in {} by values from input.
 *
 * @example
 *   service.inputToURL('/assets/{assetId}', { assetId: 123 });
 *   // => '/asset/123'
 *
 * @api private
 */
Service.prototype.inputToURL = function(url, input) {
  var variables = url.match(/\{[a-zA-Z]+\}/g);

  if(!util.isNull(variables)) {
    for(var i = 0; i < variables.length; i++) {
      var key = variables[i].replace(/(\{)|(\})/g, '');
      url = url.replace(variables[i], input[key]);
    }
  }

  return url;
};

MPX.Service = Service;
