var should = require('chai').should(),
    expect = require('chai').expect,
    nock = require('nock'),
    helpers = require('./helpers'),
    MPX = helpers.MPX;

describe('MPX.Request', function() {
  describe('.contructor()', function() {
    var fn;

    it('throws error when protocol other than http or https is given', function() {
      fn = function() { new MPX.Request({ protocol: 'ftp', hostname: 'samba.com', path: '/', method: 'POST' }); };
      expect(fn).to.throw(Error, 'Protocol "ftp" is not valid. Valid protocols are http,https');

      fn = function() { new MPX.Request({ protocol: 'http', hostname: 'samba.com', path: '/', method: 'POST' }); };
      expect(fn).to.not.throw(Error);

      fn = function() { new MPX.Request({ protocol: 'https', hostname: 'samba.com', path: '/', method: 'POST' }); };
      expect(fn).to.not.throw(Error);
    });

    it('throws error when hostname is not given', function() {
      fn = function() { new MPX.Request({ path: '/', method: 'POST' }); };

      expect(fn).to.throw(Error, 'Hostname must not be empty');
    });

    it('throws error when path is not given', function() {
      fn = function() { new MPX.Request({ hostname: 'samba.com', method: 'POST' }); };

      expect(fn).to.throw(Error, 'Path must not be empty');
    });

    it('throws error when unsupported method is not given', function() {
      fn = function() { new MPX.Request({ hostname: 'samba.com', path: '/', method: 'CONNECT' }); };

      expect(fn).to.throw(Error, 'Method "CONNECT" is not valid. Valid methods are GET,POST,PATCH,DELETE');
    });

    it('sets urlEncoded to true if HTTP method is POST or PUT', function() {
      var request;

      request = new MPX.Request({ protocol: 'https', hostname: 'samba.com', path: '/', method: 'POST' });
      request.urlEncoded.should.be.true;

      request = new MPX.Request({ protocol: 'https', hostname: 'samba.com', path: '/', method: 'GET' });
      request.urlEncoded.should.be.false;

      request = new MPX.Request({ protocol: 'https', hostname: 'samba.com', path: '/', method: 'PATCH' });
      request.urlEncoded.should.be.true;
    });
  });

  describe('.createError()', function() {
    it('creates an error based on a http status code', function() {
      var error = MPX.Request.createError(404);

      error.name.should.eq('ResourceNotFound');
      error.message.should.eq('Resource Not Found');
      error.statusCode.should.eq(404);
    });

    context('when unkonwn status code is given', function() {
      it('returns unknown error, but keeps the status code', function() {
        var error = MPX.Request.createError(418);

        error.name.should.eq('Unknown');
        error.message.should.eq('Unknown');
        error.statusCode.should.eq(418);
      });
    });
  });

  describe('#execute()', function() {
    var request = new MPX.Request({
      protocol: 'http',
      hostname: 'ah-api.mediapeers.com',
      path: '/assets',
      method: 'GET'
    });

    it('based on configuration, does HTTP request', function(done) {
      var requestData = [{
        "bucket_id": "tlv-cdd-foo",
        "key": "1w8ggxkc9qtg6bg/hard_boiled_pickpocket.xls",
        "storage": "STANDARD",
        "etag": "2160ec48c4704cc0dae958ac6b803d87",
        "size": 758120448,
        "properties": {
          "Content-Type": "application/octet-stream",
          "x-amz-meta-uploader-email": "carmine@ritchie.biz"
        },
        "deleted_at": null,
        "created_at": "2013-10-17T14:37:47.795Z",
        "updated_at": "2013-10-18T14:37:47.806Z",
        "id": "526147bbc3b5a0c2820002b3"
      }];

      var mock = nock('http://ah-api.mediapeers.com').get('/assets?format=json').reply(200, requestData);

      request.execute(function(error, data, response) {
        expect(JSON.stringify(requestData)).to.eq(JSON.stringify(data));

        mock.done();
        done();
      });
    });

    it('throws error if callback is not defined', function() {
      var fn = function() { request.execute(); };

      expect(fn).to.throw(Error, 'Callback is not a function');
    });

    describe('when 400 or 500 series errors occur', function() {
      it('returns an error', function(done) {
        var mock = nock('http://ah-api.mediapeers.com').get('/assets?format=json').reply(401, '');

        request.execute(function(error, data, response) {
          error.statusCode.should.eq(401);

          mock.done();
          done();
        });
      });
    });
  });
});
