var should = require('chai').should(),
    expect = require('chai').expect,
    sinon = require('sinon'),
    helpers = require('./helpers'),
    MPX = helpers.MPX;

describe('MPX.Service', function() {
  describe('constructor()', function() {
    afterEach(function() {
      MPX.config = new MPX.Config();
    });
    
    it('comes with configuration specified in MPX', function() {
      MPX.config.update({ endpoint: 'mediapeers.net', version: 'v20201201' });
      var service = new MPX.Service();
      
      service.config.endpoint.should.eq('mediapeers.net');
    });
    
    it('throws error if API version is not defined', function() {
      var fn = function() { new MPX.Service({ endpoint: 'mediapeers.net' }); };
      
      expect(fn).to.throw(Error, 'API version is not defined');
    });
    
    it('throws error if endpoint is not defined', function() {
      var fn = function() { new MPX.Service({ version: 'v20201201' }); };
      
      expect(fn).to.throw(Error, 'Endpoint is not defined');
    });
  });
  
  describe('defineOperation()', function() {
    var service;
    var operationName = 'listFTPDeliveries';
    var operationDefinition =  {
      http: {
        method: 'GET',
        path: '/deliveries/ftp'
      },
      input: {
        username: {
        },
        page: {
        },
        per: {
        } 
      }
    };
    
    beforeEach(function() {
      service = new MPX.Service({ endpoint: 'mediapeers.net', version: 'v20201201' });
    });
    
    it('defines a new method on service', function() {
      service.defineOperation(operationName, operationDefinition);
      expect(service[operationName]).to.be.a('function');
    });
    
    describe('defines a method that', function() {
      var request = MPX.Request;
      var callback = function() { 'callback'; };
      
      afterEach(function() {
        MPX.Request = request;
      });
      
      it('based on the definition creates new MPX.Request and executes it', function() {
        var mock = sinon.mock(MPX.Request.prototype);
        mock.expects('execute').once();
        
        MPX.Request = sinon.spy(MPX.Request);
        
        service.defineOperation(operationName, operationDefinition);
        service.listFTPDeliveries(callback);
        
        var args = {
          hostname: 'mediapeers.net',
          path: '/v20201201/deliveries/ftp',
          method: 'GET',
          data: {}
        };

        MPX.Request.calledWith(args).should.be.true;
        mock.verify();
      });
      
      it('passes only allowed parameters', function() {
        MPX.Request = sinon.spy(MPX.Request);
        
        service.defineOperation(operationName, operationDefinition);
        service.listFTPDeliveries({
          username: 'david',
          per: 200,
          invalid: 'yes'
        }, callback);
        
        MPX.Request.calledWith({
          hostname: 'mediapeers.net',
          path: '/v20201201/deliveries/ftp',
          method: 'GET',
          data: {
            username: 'david',
            per: 200
          }
        }).should.be.true;
      });
    });
  });
  
  describe('sanitizeInput()', function() {
    var service = new MPX.Service({ endpoint: 'mediapeers.net', version: 'v20201201' });
    var definition = {
      username: {},
      deliveryId: {},
      page: {}
    };
    
    it('removes values from the hash that are not defined in definition', function() {
      var input = {
        username: 'david',
        somethingExtra: 'nothing'
      };
      
      var result = service.sanitizeInput(input, definition);
      expect(result).to.have.property('username');
      expect(result).to.not.have.property('somethingExtra');
    });
    
    it('converts param keys from CamelCase to snake_case', function() {
      var input = {
        deliveryId: "super-id"
      };
      
      var result = service.sanitizeInput(input, definition);
      expect(result).to.have.property('delivery_id');
    });
    
    describe('when object contains a nested object', function() {
      var definition = {
        ingest: {
          required: true,
          deliveryId: {
            required: true
          },
          passwordDigest: {},
          properties: {
            allowAny: true
          }
        }
      };
      
      it('converts the nested objects keys as well', function() {
        var input = {
          ingest: {
            deliveryId: 'abcd',
            passwordDigest: 'whatever'
          }
        };
      
        var result = service.sanitizeInput(input, definition);
        expect(result.ingest).to.have.property('delivery_id');
        expect(result.ingest).to.have.property('password_digest');
      });
      
      it('throws an error if defintion states a nested keys is required, but is missing', function() {
        var input = {
          ingest: {
            passwordDigest: 'whatever'
          }
        };
        
        var fn = function() { service.sanitizeInput(input, definition); };
        expect(fn).to.throw(Error, 'deliveryId is a required property');
      });
      
      describe('and definition contains allowAny: true', function() {
        it('does not modify this nested object', function() {
          var input = {
            ingest: {
              deliveryId: 'abcd',
              passwordDigest: 'whatever',
              properties: {
                'x-amz-uploader-email': 'horsak@mediapeers.com'
              }
            }
          };
          
          var result = service.sanitizeInput(input, definition);
          expect(result.ingest.properties).to.have.property('x-amz-uploader-email');
        });
      });
    });
    
    it('throws an error if definition states key is required, but is missing', function() {
      definition = {
        username: {
          required: true
        }
      };
      
      var fn = function() { service.sanitizeInput({}, definition); };
      expect(fn).to.throw(Error, 'username is a required property');
    });
  });
  
  describe('sanitizeHeaders()', function() {
    var service = new MPX.Service({ endpoint: 'mediapeers.net', version: 'v20201201' }),
        headers = {
          'x-per-page': '20',
          'x-frame-options': 'SAMEORIGIN' },
        result = service.sanitizeHeaders(headers);
    
    it('removes keys and values that are not allowed by allowedHeaders', function() {
      expect(result).to.have.property('perPage');
      expect(result).to.not.have.property('frameOptions');
    });
    
    it('removes X- from customer headers and CamelCases them', function() {
      expect(result).to.have.property('perPage');
      expect(result).to.not.have.property('x-per-page');
    });
  });
  
  describe('sanitizeOutput()', function() {
    var service = new MPX.Service({ endpoint: 'mediapeers.net', version: 'v20201201' });
    
    it('camelCases all keys in the output', function() {
      var output = {
        username: 'username',
        bucket_id: 'abcd',
        password_digest: 'whatever'
      };
      
      var result = service.sanitizeOutput(output);
      expect(result).to.have.property('bucketId');
      expect(result).to.not.have.property('bucket_id');
      expect(result).to.have.property('passwordDigest');
      expect(result).to.not.have.property('password_digest');
    });
    
    describe('when input is an object that has nested object', function() {
      it('does not camelCases it', function() {
        var output = {
          bucket_id: 'abcd',
          properties: {
            'Content-Type': 'application/octet-stream',
            'x-amz-uploader-email': 'horsak@mediapeers.com'
          }
        };
        
        var result = service.sanitizeOutput(output);
        expect(result.properties).to.have.property('Content-Type');
        expect(result.properties).to.have.property('x-amz-uploader-email');
      });
    });
    
    describe('when output is an array', function() {
      it('camelCases objects in it contains', function() {
        var output = [{
          username: 'username',
          bucket_id: 'abcd',
          password_digest: 'whatever'
        }, {
          username: 'joe',
          bucket_id: 'hfgr',
          password_digest: 'just digest'
        }];
        
        var result = service.sanitizeOutput(output);
        for(var i = 0; i < output.length; i++) {
          expect(result[i]).to.have.property('bucketId');
          expect(result[i]).to.not.have.property('bucket_id');
          expect(result[i]).to.have.property('passwordDigest');
          expect(result[i]).to.not.have.property('password_digest');
        }
      });
    });
  });
  
  describe('inputToURL', function() {
    var service = new MPX.Service({ endpoint: 'mediapeers.net', version: 'v20201201' });
    
    it('substitues all variables in URL by those in input', function() {
      var url, input, result;
      
      url = '/assets/{assetId}';
      input = { assetId: 'xyz' };
      result = service.inputToURL(url, input);
      result.should.eq('/assets/xyz');
      
      url = '/deliveries/ftp/{ftpId}/assets/{assetId}';
      input = { ftpId: 1234, assetId: 'posh' };
      result = service.inputToURL(url, input);
      result.should.eq('/deliveries/ftp/1234/assets/posh');
    });
  });
});
